#include <sys/systeminfo.h>
#include <sys/types.h>
#include <time.h>
#include "xkernel.h"
#include "tcp.h"
#include "ip.h"
#include "x_stdio.h"

#define argPrefix(str) ((! strncmp(arg, str, strlen(str))) && strlen(arg)>strlen(str))
#define argEq(str) (! strcmp(arg, str) )
#define HOST_TYPE IPhost
#define FAILURE_LIMIT 2
#define DOUBLEQUOTEWRAP(x) #x
#define STRINGIFY(z) DOUBLEQUOTEWRAP(z)
#define sscanf1 sscanf
#define BUFFER_SIZE (8*1024)
#define NUM_TESTS  1
#define FILENAME "test"
#define MSG_SIZE 1024
#define CLOSE_AFTER_TEST 0

#ifndef STR2HOST
#  define STR2HOST str2ipHost
#endif

#define TRACE_VAR tracetcptestp
#define PROT_STRING "tcp"
#define TIMES 1
#define DELAY 3

typedef struct {
    XObj	lls;
    /* 
     * Client only 
     */
    XObj	controlSessn;
    int		sentMsgLength;
    Msg		savedMsg;
    /* 
     * Async only
     */
    int			clientRcvd;
    xmsg_handle_t	clientPushResult;
    xmsg_handle_t	serverPushResult;
    int			idle;
    /* 
     * stream only
     */
    int			receivedLength;
} PState;


static HOST_TYPE ServerAddr = { 0, 0, 0, 0 };
static HOST_TYPE ClientAddr = { 0, 0, 0, 0 };
static long	serverPort = 1789;

static	int	serverParam, clientParam, testsSpecified, myTestSpecified;
static	char	*serverString;
int     TRACE_VAR;

static	void 	client( Event, void * );
static 	void 	server( Event, void * );
static xkern_return_t	clientDemux( XObj, XObj, Msg * );
static 	int	runTest( XObj, int, int );
static	HOST_TYPE 	myHost;
static xkern_return_t	serverDemux( XObj, XObj, Msg * );
static  void    processOptions(void);
static xkern_return_t   closeDone( XObj );
static xkern_return_t   customClientDemux( XObj, XObj, XObj );
static xkern_return_t   customServerDemux( XObj, XObj, XObj );
static xkern_return_t   customOpenDone( XObj, XObj, XObj, XObj );
static void             printMessage (Msg *);
void printBuf ( char *ptr, long len );

int tcptest_init (XObj self);

int tcptest_init (XObj self)
{
  XObj	llp;
  PState	*ps;

  processOptions();
  if ( testsSpecified && ! myTestSpecified ) 
    {
      xTrace1(prottest, TR_SOFT_ERRORS, "Parameters indicate %s test should not run",
	      PROT_STRING);
      return 0;
    }
  
  printf("%s timing test\n", PROT_STRING);
  llp = xGetDown(self, 0);
  if ( ! xIsProtocol(llp) ) 
    {
      Kabort("Test protocol has no lower protocol");
    }
  xControl(xGetDown(self, 0), GETMYHOST, (char *)&myHost, sizeof(HOST_TYPE));
  ps = X_NEW(PState);
  bzero((char *)ps, sizeof(PState));
  self->state = (VOID *)ps;

  if (serverParam) 
    evDetach( evSchedule(server, self, 0) );
  else if (clientParam) 
    {
      STR2HOST(&ServerAddr, serverString);
      ClientAddr = myHost;
      evDetach( evSchedule(client, self, 0) );
    }
  
  return 0;
}

static void
processOptions()
{
  int		i;
  char 	*arg;
  
  for (i=1; i < globalArgc; i++) 
    {
      arg = globalArgv[i];
      if ( argEq("-s") ) 
	{
	  serverParam = 1;
	} 
      else if ( argPrefix("-c") ) 
	{
	  clientParam = 1;
	  serverString = arg + 2;
	} 
      else if ( argEq("-c")) 
	{
	  clientParam = 1;
	  serverString = globalArgv[i+1];
	  i++;
	} 
      else if ( argEq("-p")) 
	{
	  serverPort = atol(globalArgv[i+1]);
	  i++;
	} 
      else if ( argPrefix("-test")) 
	{
	  testsSpecified = 1;
	  arg += strlen("-test");
	  if ( argEq(PROT_STRING) ) 
	    {
	      myTestSpecified = 1;
	    }
	}
    }
}



/*
 *  Server startup routine
 */
static void
server( Event ev, VOID *foo )
{
  Part p;
  XObj myProtl = (XObj)foo;
  char    buf[256];
  time_t  t = time(0);
  
  printf("%s", ctime(&t));
  sysinfo(SI_HOSTNAME, buf, 255);
  printf("I am the server, running on %s\n\n", buf);
  xAssert(xIsProtocol(myProtl));

  /*
   * Since we are a test protocol, need to fill out own method table.
   */
  myProtl->demux = serverDemux;
  myProtl->opendone = customOpenDone;
  myProtl->closedone = closeDone;

  partInit(&p, 1);
  partPush(p, ANY_HOST, 0);
  partPush(p, &serverPort, sizeof(long));

  if ( xOpenEnable(myProtl, myProtl, xGetDown(myProtl, 0), &p)
      == XK_FAILURE ) 
    {
      printf("%s test server can't openenable lower protocol\n",
	     PROT_STRING);
    } 
  else 
    {
      printf("%s test server done with xopenenable\n", PROT_STRING);
    }
  return;
}


/*
 * Client initialization routine 
 */
static void
client( Event ev, VOID *foo )
{
  XObj	myProtl = (XObj)foo;
  PState	*ps;
  char    buf[256];
  time_t  t = time(0);
  int i;
  
  printf("%s", ctime(&t));
  sysinfo(SI_HOSTNAME, buf, 255);
  printf("I am the client, running on %s\n", buf);

  xAssert(xIsXObj(myProtl));
  ps = (PState *)myProtl->state;

  /*
   * Since we are a test protocol, need to fill out own method table.
   */
  myProtl->demux = clientDemux;

  for (i = 0; i < NUM_TESTS; i++)
    runTest(myProtl, 0, 0);

  printf("End of test\n");
  fprintf(stderr, "End of test\n");
}


static
#if defined(__GNUC__)
inline
#endif
void checkHandle( h, str )
    xmsg_handle_t	h;
    char		*str;
{
    switch ( h ) {
      case XMSG_ERR_HANDLE:
      case XMSG_ERR_WOULDBLOCK:
	sprintf(errBuf, "%s returns error handle %d", str, h);
	Kabort(errBuf);
      default:
	;
    }
}


int runTest( self, len, testNumber )
    XObj	self;
    int 	len;
    int		testNumber;
{
  Msg	        *msg;
  char	        *buf;
  PState	*ps = (PState *)self->state;
  FILE          *f;
  size_t        bytes = MSG_SIZE;
  int           c = 1;
  Part	p[2];
  
  partInit(p, 1);
  partPush(p[0], &ServerAddr, sizeof(IPhost));
  partPush(p[0], &serverPort, sizeof(long));

  ps->lls = xOpen(self, self, xGetDown(self, 0), p);
  if ( ps->lls == ERR_XOBJ ) 
    {
      printf("%s test: open failed!\n", PROT_STRING);
      Kabort("End of test");
    }

  customOpenDone(self, ps->lls, xGetDown(self, 0), self);

  xAssert(ps);

  f = fopen (FILENAME, "r");

  while (bytes == MSG_SIZE)
    {
      msg = X_NEW (Msg);
      
      msgConstructAllocate(msg, MSG_SIZE, (char **)&buf);

      bytes = fread ((void *) buf, 1, MSG_SIZE, f);

      if (bytes != MSG_SIZE)
	{
	  msgTruncate (msg, bytes);
	}

      ps->clientPushResult = xPush(ps->lls, msg);
      checkHandle(ps->clientPushResult, "client push");

      xIfTrace(prottest, TR_MAJOR_EVENTS) 
	{
	  putchar('.');
	  if (! (c++ % 50)) 
	    {
	      putchar('\n');
	    }
	}

      msgDestroy(msg);
      free (msg);
    }


  xControl (ps->lls, TCP_PUSH, 0, 0);
  
  fclose (f);

  xClose (ps->lls);
  
  Delay (10000);
  
  return 0;
}


void printBuf ( char *ptr, long len )
{
  char buf[len + 1];

  memcpy (buf, ptr, len);

  buf[len] = '\0';
  printf("%s", buf);
}


static void printMessage (dg)
     Msg *dg;
{
  MsgIterator ctx;
  long len;
  char *src;

  if (msgLen (dg) == 0)
    return;

  msgIterateInit(&ctx, dg);
  while ((src = msgIterateNext(&ctx, &len)) != 0) 
    {
      printBuf (src, len);
    }

  msgIterateDone(&ctx);
}


xkern_return_t
serverDemux( self, lls, dg )
    XObj self, lls;
    Msg *dg;
{
  static int c = 1;
  
  printMessage (dg);

/*
  xIfTrace(prottest, TR_MAJOR_EVENTS) 
    {
      putchar('.');
      if (! (c++ % 50)) 
	{
	  putchar('\n');
	}
    }
*/
  
  customServerDemux (self, lls, 0);
  
  return XK_SUCCESS;
}



static xkern_return_t
clientDemux( self, lls, dg )
    XObj self, lls;
    Msg *dg;
{
  /*
   * In this test, the server doesn't send back any messages, so 
   * this routine is never called.
   */

  return XK_SUCCESS;
}


static xkern_return_t
closeDone( lls )
    XObj	lls;
{
  xTrace2(prottest, TR_MAJOR_EVENTS, "%s test -- closedone (%x) called",
	  PROT_STRING, lls);
  if ( serverParam ) 
    xClose(lls);
  return XK_SUCCESS;
}


static xkern_return_t
customOpenDone( self, s, llp, hlpType )
    XObj self, s, llp, hlpType;
{
  u_short space = BUFFER_SIZE;

  xTrace0(prottest, TR_MAJOR_EVENTS, "tcp test program openDone");

  if (serverParam)
    xDuplicate(s);

  if (xControl(s, TCP_SETRCVBUFSIZE, (char*)&space, sizeof(space)) < 0) 
    xError("saveServerSessn: TCP_SETRCVBUFSIZE failed");

  if (xControl(s, TCP_SETRCVBUFSPACE, (char*)&space, sizeof(space)) < 0)
    xError("saveServerSessn: TCP_SETRCVBUFSPACE failed");

  return XK_SUCCESS;
}


static xkern_return_t
customClientDemux( self, lls, dg )
    XObj	self, lls, dg;
{
  u_short space = BUFFER_SIZE;

  xTrace1(prottest, TR_DETAILED, "TCP custom demux resets buffer to size %d",
	  BUFFER_SIZE);
  if (xControl(lls, TCP_SETRCVBUFSPACE, (char*)&space, sizeof(space)) < 0) 
    xError("TCP custom test_demux: TCP_SETRCVBUFSPACE failed");
  
  return XK_SUCCESS;
}


static xkern_return_t
customServerDemux( self, lls, dg )
    XObj	self, lls, dg;
{
    return customClientDemux( self, lls, dg );
}
